// SPDX-License-Identifier: MIT

#[cfg(test)]
mod cache_info;
#[cfg(test)]
mod expires;
#[cfg(test)]
mod loopback;
#[cfg(test)]
mod mpls;
#[cfg(test)]
mod multipath;
#[cfg(test)]
mod realm;
#[cfg(test)]
mod route_flags;
#[cfg(test)]
mod uid;
#[cfg(test)]
mod via;
